#ifndef FrameCPP__Common__Dictionary_HH
#define FrameCPP__Common__Dictionary_HH

#include <map>
#include <iostream>
#include <typeinfo>

#include "general/gpstime.hh"
#include "general/unordered_map.hh"

#include "framecpp/Common/FrameSpec.hh"
#include "framecpp/Common/StreamRefInterface.hh"

#define HASH_STREAMREF_BY_ID 1

namespace FrameCPP
{
  namespace Common
  {
    class OStream;
    class OFrameStream;

    //-------------------------------------------------------------------
    /// \brief Keep track of structures written to or read from stream
    //-------------------------------------------------------------------
    class Dictionary
    {
    public:
      typedef FrameSpec::Object object_type;
      typedef StreamRefInterface streamref_type;
      typedef std::pair< StreamRefInterface::class_type,
			 StreamRefInterface::instance_type >
      ptr_struct_key_type;


      Dictionary( );

      ~Dictionary( );

      void Clear( );

      const streamref_type* Ref( const ptr_struct_key_type& Key ) const;
      const object_type* Ref( const streamref_type* SR ) const;
      streamref_type* Ref( const object_type* ) const;

      void Ref( streamref_type* SR, const object_type* );

      /// \brief Create a reference
      streamref_type* RefCreate( const object_type* Obj,
				 const OStream& Stream );

      void Remove( const ptr_struct_key_type& Key );

      void Remove( const object_type* Key );

      void ResetInstanceCounts( );

      void SetInstance( streamref_type& StreamRef );

      void Next( const object_type* Obj,
		 const object_type* Next );

      const object_type* Next( const object_type* Obj ) const;

    private:
      friend class FrameStream;

#if HASH_STREAMREF_BY_ID
      struct ptr_struct_hash
      {
	inline size_t operator( )( const ptr_struct_key_type& Key ) const
	{
#if SIZEOF_SIZE_T == 4
	  register size_t retval = size_t( (Key.first << 24) | Key.second );
	return retval;
#else
	return ( ( size_t( Key.second ) << 8 ) | Key.first );
#endif
	}

	inline bool operator()( const ptr_struct_key_type& LHS,
				const ptr_struct_key_type& RHS ) const
	{
	  return ( ( LHS.first == RHS.first )
		   && ( LHS.second == RHS.second ) );
	}
      };
#endif /* HASH_STREAMREF_BY_ID */

      typedef General::unordered_map< const object_type*, streamref_type*,
				      General::hash_pointer< const object_type > >
      streamref_container_type;

      typedef General::unordered_map< const streamref_type*, const object_type*,
				      General::hash_pointer< const streamref_type > >
      object_container_type;

      typedef General::unordered_map< const object_type*, bool,
				      General::hash_pointer< const object_type > >
      ownership_container_type;

      typedef General::unordered_map< FrameSpec::class_type,
				      FrameSpec::instance_type >
      instance_count_container_type;

      typedef General::unordered_map< const FrameSpec::Object*,
				      const FrameSpec::Object*,
				      General::hash< const void* > >
      next_container_type;

#if HASH_STREAMREF_BY_ID
      typedef General::unordered_map< ptr_struct_key_type,
				      const streamref_type*,
				      ptr_struct_hash,
				      ptr_struct_hash >
      streamref_by_id_container_type;
#else
      typedef std::map< ptr_struct_key_type,
			const streamref_type* >
      streamref_by_id_container_type;
#endif

      const streamref_type*		m_stream_ref;

      streamref_container_type		m_streamref_container;
      object_container_type		m_object_container;
      mutable ownership_container_type	m_ownership_container;
      instance_count_container_type	m_instance_count_container;
      next_container_type		m_next_container;
      streamref_by_id_container_type	m_streamref_by_id_container;
    };

    inline void Dictionary::
    Clear( )
    {
      m_instance_count_container.clear( );
      m_streamref_by_id_container.clear( );
      m_object_container.clear( );
      for ( streamref_container_type::const_iterator
	      cur = m_streamref_container.begin( ),
	      last = m_streamref_container.end( );
	    cur != last;
	    ++cur )
      {
	if ( m_ownership_container[ cur->first ] )
	{
	  delete cur->first;
	}
	delete cur->second;
      }
      m_streamref_container.clear( );
      m_ownership_container.clear( );
    }

    inline void Dictionary::
    Next( const FrameSpec::Object* Obj,
	  const FrameSpec::Object* Next )
    {
      m_next_container[ Obj ] = Next;
    }

    inline const FrameSpec::Object* Dictionary::
    Next( const FrameSpec::Object* Obj ) const
    {
      const FrameSpec::Object* retval = (FrameSpec::Object*)NULL;
      next_container_type::const_iterator i = m_next_container.find( Obj );

      if ( i != m_next_container.end( ) )
      {
	retval = i->second;
      }

      return retval;
    }


    inline const Dictionary::streamref_type* Dictionary::
    Ref( const ptr_struct_key_type& Key ) const
    {
      streamref_by_id_container_type::const_iterator retval
	= m_streamref_by_id_container.find( Key );
      if ( retval == m_streamref_by_id_container.end( ) )
      {
	return (const streamref_type*)NULL;
      }
      return retval->second;
    }

    inline const Dictionary::object_type* Dictionary::
    Ref( const streamref_type* SR ) const
    {
      object_container_type::const_iterator retval
	= m_object_container.find( SR );
      if ( retval == m_object_container.end( ) )
      {
	return (const object_type*)NULL;
      }
      m_ownership_container[ retval->second ] = false;
      return retval->second;
    }

    inline Dictionary::streamref_type* Dictionary::
    Ref( const object_type* Object ) const
    {
      streamref_container_type::const_iterator retval
	= m_streamref_container.find( Object );
      if ( retval == m_streamref_container.end( ) )
      {
	return (streamref_type*)NULL;
      }
      return retval->second;
    }

    inline void Dictionary::
    Ref( streamref_type* SR, const object_type* Obj )
    {
      //-----------------------------------------------------------------
      if ( SR )
      {
	ptr_struct_key_type p( SR->GetClass( ), SR->GetInstance( ) );

	m_streamref_by_id_container[ p ] = SR;
      }
      //-----------------------------------------------------------------
      // :TODO: Need to handle creating SR if SR is null
      m_streamref_container[ Obj ] = SR;
      m_object_container[ SR ] = Obj;
      m_ownership_container[ Obj ] = true;
    }

    //-------------------------------------------------------------------
    /// This method is used for creating a reference that would be
    ///   appropriate for writing to a stream.
    //-------------------------------------------------------------------
    inline Dictionary::streamref_type* Dictionary::
    RefCreate( const object_type* Obj, const OStream& Stream )
    {
      streamref_type* retval = Ref( Obj );

      register INT_4U c;

      if ( ( retval == (streamref_type*)NULL )
	   && ( ( c = Obj->GetClass( ) ) > 0 ) )
      {
	retval = dynamic_cast< streamref_type* >( m_stream_ref->Clone( ) );
	if ( retval )
	{
	  retval->SetClass( c );
	  retval->SetInstance( m_instance_count_container[ c ]++ );
	  Ref( retval, Obj );
	  m_ownership_container[ Obj ] = false;
	}
      }
      return retval;
    }

    inline void Dictionary::
    Remove( const ptr_struct_key_type& Key )
    {
      streamref_by_id_container_type::iterator sridi
	= m_streamref_by_id_container.find( Key );
      if ( sridi == m_streamref_by_id_container.end( ) )
      {
	return;
      }
      object_container_type::const_iterator
	oci = m_object_container.find( sridi->second );
      if ( oci != m_object_container.end( ) )
      {
	delete oci->first;
	m_streamref_container.erase( oci->second );
      }
      m_object_container.erase( sridi->second );
      m_ownership_container.erase( sridi->second );
      m_streamref_by_id_container.erase( sridi );
    }

    inline void Dictionary::
    Remove( const object_type* Key )
    {
      streamref_container_type::iterator
	srci = m_streamref_container.find( Key );
      if ( srci != m_streamref_container.end( ) )
      {
	//----------------------------------------------------------------
	// Remove from secondary lists
	//----------------------------------------------------------------
	m_ownership_container.erase( Key );
	if ( srci->second )
	{
	  ptr_struct_key_type p( srci->second->GetClass( ),
				 srci->second->GetInstance( ) );
	
	  m_streamref_by_id_container.erase( p );

	  m_object_container.erase( srci->second );
	}

	//----------------------------------------------------------------
	// Take out of list
	//----------------------------------------------------------------
	delete srci->second;
	m_streamref_container.erase( srci );
      }
    }

    inline void Dictionary::
    ResetInstanceCounts( )
    {
      Clear( );
    }
  } // namespace - Common
} // namespace - FrameCPP
#endif /* FrameCPP__Common__Dictionary_HH */
