#include "framecpp/config.h"

#include <iostream>

#include "framecpp/Common/IOStream.hh"
#include "framecpp/Common/FrameStream.hh"
#include "framecpp/Common/CheckSumFilter.hh"
#include "framecpp/Common/Description.hh"
#include "framecpp/Common/Verify.hh"

#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"

#include "framecpp/Version6/FrameSpec.hh"

using namespace FrameCPP::Version_6;
using FrameCPP::Common::CheckSum;
using FrameCPP::Common::CheckSumFilter;
using FrameCPP::Common::Description;
using FrameCPP::Common::FrameSpec;
using FrameCPP::Common::IStream;
using FrameCPP::Common::OStream;
using FrameCPP::Common::OFrameStream;


FrEndOfFile::
FrEndOfFile( )
  : Common::FrEndOfFile( StructDescription( ) ),
    m_nFrames( 0 ),
    m_nBytes( 0 ),
    m_chkType( 0 ),
    m_chkSum( 0 ),
    m_seekTOC( 0 )
{
}

FrEndOfFile::
FrEndOfFile( IStream& Stream )
  : Common::FrEndOfFile( StructDescription( ) )
{
  //---------------------------------------------------------------------
  // Read data stream
  //---------------------------------------------------------------------
  Stream >> m_nFrames
	 >> m_nBytes
	 >> m_chkType
    ;
  //---------------------------------------------------------------------
  // Check for checksum calculations
  //---------------------------------------------------------------------
  CheckSumFilter*	crc = Stream.GetCheckSumFile( );

  if ( crc )
  {
   static const chkSum_type		chkSum = 0;

    crc->Filter( &chkSum, sizeof( chkSum ) );	// Use 0 value
    Stream.rdbuf( )->FilterRemove( crc );
  }
  Stream >> m_chkSum
    ;
  if ( crc )
  {
    Stream.rdbuf( )->FilterAdd( crc );
  }
  Stream >> m_seekTOC
    ;
  if ( crc )
  {
    crc->Value( );	// Stop processing
  }
}

FrEndOfFile::
~FrEndOfFile( )
{
}

FrameSpec::size_type FrEndOfFile::
Bytes( const Common::StreamBase& Stream ) const
{
  return
    sizeof( m_nFrames )
    + sizeof( m_nBytes )
    + sizeof( m_chkType )
    + sizeof( m_chkSum )
    + sizeof( m_seekTOC )
    ;
}

FrEndOfFile* FrEndOfFile::
Create( Common::IStream& Stream ) const
{
  return new FrEndOfFile( Stream );
}

const char* const FrEndOfFile::
ObjectStructName( ) const
{
  return StructName( );
}

const Description* FrEndOfFile::
StructDescription( )
{
  static Description ret;

  if ( ret.size( ) == 0 )
  {
    ret( FrSH( "FrEndOfFile", FrEndOfFile::s_object_id,
	       "End of File Data Structure" ) );
    ret( FrSE( "nFrames", "INT_4U",
	       "Number of frames in this file") );
    ret( FrSE( "nBytes", "INT_8U",
	       "Total number of bytes in this file"
	       " ( 0 if NOT computed )") );
    ret( FrSE( "chkType", "INT_4U",
	       "Checksum schemes: 0=none, 1=CRC, >2 presently unsued" ) );
    ret( FrSE ( "chkSum", "INT_4U", "File checksum." ) );
    ret( FrSE ( "seekTOC", "INT_8U",
		"Byes to back up to the beginning of the table of"
		" contents structure. If seekTOC == 0, then there"
		" is no TOC for this file." ) );
  }

  return &ret;
}

void FrEndOfFile::
VerifyObject( Common::Verify& Verifier,
	      Common::IFrameStream& Stream ) const
{
  Verifier.SeenFrEndOfFile( true );
  Verifier.ExamineFrEndOfFileChecksum( Stream, GetChkType( ), m_chkSum );
}
		  
void FrEndOfFile::
Write( OStream& Stream ) const
{
  //---------------------------------------------------------------------
  // Get local information
  //---------------------------------------------------------------------
 
  const nFrames_type	nFrames( Stream.GetNumberOfFrames( ) );
  const nBytes_type	nBytes( Stream.tellp( )
				+ std::streampos( Bytes( Stream ) ) );
  const seekTOC_type	seekTOC( ( Stream.GetTOCOffset( ) > 0 )
				 ? nBytes - Stream.GetTOCOffset( )
				 : 0 );

  chkType_type		chkType = CheckSum::NONE;
  chkSum_type		chkSum = 0;

  //---------------------------------------------------------------------
  // Calculate checksum
  //---------------------------------------------------------------------
  CheckSumFilter*
    crc = Stream.GetCheckSumFile( );

  if ( crc )
  {
    chkType = crc->Type( );

    crc->Filter( &nFrames, sizeof( nFrames ) );
    crc->Filter( &nBytes, sizeof( nBytes ) );
    crc->Filter( &chkType, sizeof( chkType ) );
    crc->Filter( &chkSum, sizeof( chkSum ) );
    crc->Filter( &seekTOC, sizeof( seekTOC ) );
    
    chkSum = crc->Value( );
    
    Stream.SetCheckSumFile( CheckSum::NONE );
  }

  //---------------------------------------------------------------------
  // Write out to the stream
  //---------------------------------------------------------------------
  Stream << nFrames
	 << nBytes
	 << chkType
	 << chkSum
	 << seekTOC
    ;
}

bool FrEndOfFile::
operator==( const FrEndOfFile& RHS ) const
{
  return ( ( m_nFrames == RHS.m_nFrames )
	   && ( m_nBytes == RHS.m_nBytes )
	   && ( m_chkType == RHS.m_chkType )
	   && ( m_chkSum == RHS.m_chkSum )
	   && ( m_seekTOC == RHS.m_seekTOC ) );
}

bool FrEndOfFile::
operator==( const Common::FrameSpec::Object& Obj ) const
{
  return ( *this == Obj );
}

std::auto_ptr< FrameSpec::Object > FrEndOfFile::
demote( INT_2U Target, IStream* Stream ) const
{
  throw
    Unimplemented( "Object* FrEndOfFile::demote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}

std::auto_ptr< FrameSpec::Object > FrEndOfFile::
promote( INT_2U Target, Object* Obj, IStream* Stream ) const
{
  throw
    Unimplemented( "Object* FrEndOfFile::promote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}
