#include "general/config.h"

#include <sys/types.h>

#include <dirent.h>   // for DIR*
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <stdexcept>
#include <sstream>

#include "general/Directory.hh"
#include "general/System.hh"

#ifndef NAME_MAX
#ifdef FILENAME_MAX
#define NAME_MAX FILENAME_MAX
#endif
#endif

namespace General
{
  class Directory::Internals
  {
  private:
    typedef struct dirent dirent_type;

  public:
    Internals( const std::string& DirName, bool OpenDirectory )
    try
      : m_dirname( DirName ),
	m_dirp( (DIR*)NULL ),
	m_fd( -1 ),
	m_buffer( (dirent_type*)NULL )
    {
      if ( OpenDirectory )
      {
	Open( );
      }
    }
    catch( ... )
    {
      cleanup( );
      throw;	// Rethrow exception
    }

    ~Internals( )
    {
      cleanup( );
    }

    inline void
    Close( )
    {
      if ( m_dirp )
      {
	closedir( m_dirp );
	m_dirp = (DIR*)NULL;
      }
    }

    inline const std::string&
    DirName( ) const
    {
      return m_dirname;
    }

    inline int
    Fd( ) const
    {
      return m_fd;
    }

    inline const char* const
    Filename( ) const
    {
      return m_buffer->d_name;
    }

    inline bool
    Next( )
    {
      dirent_type*	result;
      int error = readdir_r( m_dirp, m_buffer, &result );
      return ( ( error == 0 ) && ( result != (dirent_type*)NULL ) );
    }

    void
    Open( )
    {
      if ( m_dirp )
      {
	rewinddir( m_dirp );
	return;
      }

#if HAVE_DIRFD
      m_dirp = opendir( m_dirname.c_str() );
#elif HAVE_FDOPENDIR /* HAVE_DIRFD */
      m_fd = open( m_dirname.c_str( ), O_RDONLY );
      if ( m_fd >= 0 )
      {
	m_dirp = fdopendir( m_fd );
      }
#else
#error Unable to obtain the directory file descriptor
#endif /* HAVE_DIRFD */
   
      if( m_dirp == 0 )
      {
	std::ostringstream	msg;

	msg << "General::Directory::Internals::Internals( ): "
	    << "opendir( ) failed for '" << m_dirname << "'"
	    << General::System::ErrnoMessage( );
   
	throw std::runtime_error( msg.str( ) );
      }
#if HAVE_DIRFD
      m_fd = dirfd( m_dirp );
#endif /* HAVE_DIRFD */
      //-----------------------------------------------------------------
      // Ensure that we start reading at the start of the directory
      //-----------------------------------------------------------------
      rewinddir( m_dirp );

      // Allocate the buffer only once.
      if ( ! m_buffer )
      {
	//---------------------------------------------------------------
	// Need to find out the size of the buffer that needs to be
	//    allocated.
	//---------------------------------------------------------------
	long name_max = -1;
#   if defined(HAVE_FPATHCONF) && defined(_PC_NAME_MAX)
	name_max = fpathconf( m_fd, _PC_NAME_MAX);
	if (name_max == -1)
	{
#     if defined(NAME_MAX)
	  name_max = NAME_MAX;
#     else
	  std::ostringstream	msg;

	  msg << "General::Directory::Internals::Internals( ): "
	      << "Unable to allocate directory buffer";

	  throw std::runtime_error( msg.str( ) );
#     endif /* defined(NAME_MAX) */
	}
#   else
#       if defined(NAME_MAX)
	name_max = NAME_MAX;
#       else
#           error "buffer size for readdir_r cannot be determined"
#       endif
#   endif
	size_t struct_size
	  = (size_t)offsetof(dirent_type, d_name) + name_max + 1;
	//---------------------------------------------------------------
	// After finding the size of the buffer, it is time to allocate
	//   the memory for the buffer for reading directory information.
	//---------------------------------------------------------------
	m_buffer = (dirent_type*)malloc( struct_size );
      }
    }

  private:
    std::string		m_dirname;
    DIR*		m_dirp;
    int			m_fd;
    dirent_type*	m_buffer;

    void cleanup( )
    {
      m_fd = -1;
      try
      {
	Close( );
      }
      catch( ... )
      {
      }
      if ( m_buffer )
      {
	free( m_buffer );
	m_buffer = (dirent_type*)NULL;
      }
    }
  };

  Directory::
  Directory( const std::string& DirName, bool OpenDirectory )
    : m_internals( (Internals*)NULL )
  {
    m_internals = new Internals( DirName, false );
    if ( OpenDirectory )
    {
      try
      {
	Open( );
      }
      catch( ... )
      {
      }
    }
  }

  Directory::
  ~Directory( )
  {
    delete m_internals;
  }

  int Directory::
  Fd( ) const
  {
    return m_internals->Fd( );
  }

  void Directory::
  Close( )
  {
    return m_internals->Close( );
  }

  bool Directory::
  Next( )
  {
    return m_internals->Next( );
  }

  const char* const Directory::
  EntryName( ) const
  {
    return m_internals->Filename( );
  }

  const std::string& Directory::
  Name( ) const
  {
    return m_internals->DirName( );
  }

  void Directory::
  Open( )
  {
    return m_internals->Open( );
  }
}
